
/************************************************************************
 *																								*
 *		D51 8051/52 Disassembler - Copyright (C) 1995 by						*
 *		Jeffery L. Post																	*
 *		22726 Benner Ave.																	*
 *		Torrance, CA  90505																*
 *																								*
 *		D51P1.C - Pass one of disassembly											*
 *																								*
 *		Examine opcodes for internal references to other memory				*
 *		locations. If such references are found, flag the referenced		*
 *		location so that a label can be generated in the output file		*
 *		during pass two.																	*
 *																								*
 *  Version 2.1 - 10/21/95																*
 *																								*
 *	This program is free software; you can redistribute it and/or modify	*
 *	it under the terms of the GNU General Public License as published by	*
 *	the Free Software Foundation; either version 2 of the License, or		*
 *	(at your option) any later version.												*
 *																								*
 *	This program is distributed in the hope that it will be useful,		*
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of			*
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the			*
 *	GNU General Public License for more details.									*
 *																								*
 *	You should have received a copy of the GNU General Public License		*
 *	along with this program; if not, write to the Free Software				*
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.				*
 *																								*
 ************************************************************************/

#include	<stdio.h>
#include	<dos.h>
#include	<ctype.h>
#include	<alloc.h>
#include	"d51.h"

#define	SYMBOL	0
#define	LABEL		1

/*********************
 *							*
 *		Prototypes		*
 *							*
 *********************/

void	pass0(void);
void	pass1(void);
char	*get_adrs(char *text, word *val);
void	dump_ascii(word adrs);
void	dump_bytes(word adrs);

extern char	*find_entry(word val, struct sym *ptr);
extern void	add_entry(word val, char *symbol, int type);

#ifdef	DEBUG

void	dump_pflg(void);

extern void	dump_sym(void);
extern void	dump_lab(void);

#endif

/*********************
 *							*
 *	Global variables	*
 *							*
 *********************/

extern FILE	*fp;						/* dos file struct				*/
extern int	hexflag;					/* append hex flag				*/
extern int	kcnt;						/* output char counter			*/
extern word	himark;					/* highest data adrs				*/
extern word	offset;					/* program offset					*/
extern byte	far *pmem;				/* program data pointer			*/
extern byte	far *pflg;				/* pointer to program flags	*/
extern char	string[ASCLIMIT];		/* ascii data for defb			*/
extern word	asccount;				/* count for string data		*/
extern word	byte_cnt;				/* count for byte binary data	*/
extern byte	dump;						/* dump just done flag			*/
extern char	defbstr[8];				/* string for defined bytes	*/
extern char	ascistr[8];				/* string for defined ascii	*/
extern byte opttbl[256];			/* code option bit flags		*/
extern char	ctl[32];					/* control file name				*/
extern char	linebuffer[128];		/* input line buffer				*/
extern struct sfrent sfr[256];	/* SFR names						*/
extern struct sfrent sfrbits[128]; /* SFR bit names				*/
extern struct sfrent rbname[32];	/* register bank names			*/
extern struct sym	*sym_tab;		/* symbol table pointer			*/
extern struct sym	*lab_tab;		/* label table pointer			*/

/************************************************************************
 *																								*
 *		Read control file, if it exists, and flag areas as code, text,		*
 *		or whatever. Also handle sfr name changes, labels, etc.				*
 *																								*
 ************************************************************************/

void pass0(void)
{
	char	*text, func, c, *ltext;
	word	start, stop, code, temp;

	int	i;

	fp = fopen(ctl, "r");
	if (fp != NULL)			/* if control file exists... */
	{
		printf("\nReading control file");
		while (!feof(fp))						/* until end of file... */
		{
			start = stop = 0;
			*linebuffer = '\0';				/* clear previous line */
			fgets(linebuffer, 127, fp);	/* read one line */
			text = &linebuffer[1];
			text = get_adrs(text, &start);
			while (1)
			{
				c = *text++;
				if (c != ' ' && c != '\t')	/* skip whitespace */
					break;
			}
			if (c == '\n' || c == ';')		/* if only one numeric... */
				--text;							/* back up to newline */
			func = c;							/* save operator */
			ltext = text;
			--ltext;
			text = get_adrs(text, &stop);
			if (func == '+')					/* check for valid operator */
				stop += (start - 1);
			else if (func == '-' && !stop)
				stop = start;

			/* for test only...
			printf("\nToken = %c, start = 0x%04x, stop = 0x%04x\n",
				linebuffer[0], start, stop);	*/

			switch (toupper(linebuffer[0]))
			{
				case 'A':								/* address */
					do
					{						/* get address to reference */
						code = ((word) pmem[start]) << 8;
						temp = pmem[start + 1] & 0xff;
						code |= temp;
						pflg[code] |= PF_REF;	/* flag referenced address */
						pflg[code] &= ~(PF_CLREF | PF_SPLIT);
						pflg[start++] = PF_ADRS;	/* set branch flags */
						pflg[start] = PF_ADRS;		/*  to address code */
						start++;
					} while (start < stop);
					break;

				case 'B':								/* byte binary */
					do
					{
						pflg[start] = PF_BYTE;
						start++;
					} while (start <= stop);
					break;

				case 'C':								/* code */
					code = start;
					do
					{
						if (pflg[start] == PF_INIT)
							pflg[start] = PF_CLREF;		/* flag as valid code data */
						else
							pflg[start] &= ~PF_NOINIT;
						start++;
					} while (start <= stop);
					break;

				case 'F':								/* modify SFR name */
					if (start < 0x80 || start > 0xff)
					{
						printf("\rInvalid SFR address: 0x%x in '%s'\n",
						linebuffer[0], linebuffer);
					}
					else
					{
						for (stop=0; stop<7; stop++)	/* transfer new name */
						{
							func = *ltext++;
							if (isalnum(func))
								sfr[start].dent[stop] = func;
							else
							{
								sfr[start].dent[stop] = '\0';	/* terminate name */
								break;
							}
						}
					}
					if (stop >= 7)
						sfr[start].dent[7] = '\0';
					break;

				case 'I':								/* ignore initialized data */
					do
					{
						pflg[start] = PF_INIT;
						start++;
					} while (start <= stop);
					break;

				case 'K':								/* modify SFR bit name */
					if (start < 0x80 || start > 0xff)
					{
						printf("\rInvalid SFR bit address: 0x%x in '%s'\n",
						linebuffer[0], linebuffer);
					}
					else
					{
						for (stop=0; stop<7; stop++)	/* transfer name */
						{
							func = *ltext++;
							if (isalnum(func))
								sfrbits[start].dent[stop] = func;
							else
							{
								sfrbits[start].dent[stop] = '\0';
								break;
							}
						}
					}
					if (stop >= 7)
						sfrbits[start].dent[7] = '\0';	/* null terminate name */
					break;

				case 'L':								/* label */
					pflg[start] |= PF_REF;			/* flag reference */
					pflg[start] &= ~(PF_CLREF | PF_SPLIT);
					if (isalnum(*ltext))
						add_entry(start, ltext, LABEL);
					break;

				case 'R':								/* modify register name */
					if (start > 31)
					{
						printf("\rInvalid register address: 0x%x in '%s'\n",
						linebuffer[0], linebuffer);
					}
					else
					{
						for (stop=0; stop<7; stop++)	/* transfer register name */
						{
							func = *ltext++;
							if (isalnum(func))
								rbname[start].dent[stop] = func;
							else
							{
								rbname[start].dent[stop] = '\0';
								break;
							}
						}
					}
					if (stop >= 7)
						rbname[start].dent[7] = '\0';	/* null terminate reg name */
					break;

				case 'S':								/* symbol */
					add_entry(start, ltext, SYMBOL);
					break;

				case 'T':								/* text */
					do
					{
						pflg[start] = PF_ASCII;
						start++;
					} while (start <= stop);
					break;

				case 'W':								/* word binary */
					do
					{
						pflg[start] = PF_WORD;
						start++;
					} while (start <= stop);
					break;

				case 0x00:								/* ignore empty lines */
				case '\n':
				case '\r':
				case '\t':
				case ' ':
				case ';':								/* ignore commented out lines */
					break;

				default:						/* somebody didn't read the docs... */
					printf("\rUnknown control code: 0x%02x in '%s'\n",
						linebuffer[0], linebuffer);
					break;
			}
		}
	}

#ifdef	DEBUG

	dump_lab();
	dump_sym();
	dump_pflg();

#endif
}

/*********************************************************
 *																			*
 *		Get hexadecimal number from line in control file.	*
 *		Return updated character pointer.						*
 *																			*
 *********************************************************/

char *get_adrs(char *text, word *val)
{
	word	result, start;
	char	c;

	result = start = 0;
	while (c = toupper(*text))
	{
		if (c == ';')			/* beginning of comment, ignore all else */
			break;
		if (c == '\n')			/* necessary because isspace() includes \n */
			break;
		if (isspace(c))		/* skip leading whitespace */
		{
			text++;
			if (start)			/* if result already begun... */
				break;
		}
		else if (!isxdigit(c))		/* done if not hexadecimal character */
			break;
		else
		{
			start = 1;			/* flag beginning of result conversion */
			c = (c > '9') ? c - 0x37 : c - 0x30;
			result <<= 4;
			result |= ((word) c & 0xf);
			text++;
		}
	}
	*val = result;				/* pass number back to caller */
	return(text);				/* and return updated text pointer */
}

#ifdef	DEBUG

/*	Diagnostic - output program flag table */

void dump_pflg(void)
{
	word	i, j, k, t;

	printf("\nProgram flags:\n0000: ");
	for (i=0, j=0, k=0; i<0xffff; i++)
	{
		printf("%02x ", pflg[i]);
		j++;
		if (j == 16)
		{
			j = 0;
			k++;
			if (k == 20)
			{
				k = 0;
				t = getch();
				if (t == 0x1b)
					return;
			}
			printf("\n%04x: ", i + 1);
		}
	}
}

#endif

/***********************************************************************

	Pass one of disassembly. Examine opcodes for internal references
	to other memory locations. If such references are found, flag the
	referenced location so that a label can be generated in the output
	file during pass two.

************************************************************************/

void pass1(void)
{
	word	i, l, pc, rel;
	byte	j, k, mask;
	char	*inp;
	int	line;

	printf("Pass 1 0000");
	for (i=offset; i<himark; )
	{
		l = i & 0xff;
		k = pmem[i];						/* get stored opcode */
		if (pflg[i] & PF_NOINIT)		/* ignore un-initialized data */
			i++;
		else if (!(pflg[i] & (PF_ADRS | PF_WORD | PF_BYTE | PF_ASCII)))
		{										/* if code... */
			mask = (byte) (PF_CLREF | PF_SPLIT);
			j = opttbl[k];					/* get option byte */

			if ((j & OPT_MASK) == OPT_LREF)	/* if ljmp, lcall, or mov dptr */
			{
				pc = ((pmem[i+1] & 0xff) << 8) | (pmem[i+2] & 0xff);
				if (pc >= offset && pc <= himark)
					mask |= PF_NOINIT;
				pflg[pc] = pflg[pc] & ~mask | PF_REF;		/* flag reference */
			}

			if (j & OPT_11)				/* if 11 bit adrs (ajmp & acall) */
			{
				pc = (i & 0xf800) | ((k & 0xe0) << 3) | (pmem[i+1] & 0xff);
				if (pc >= offset && pc <= himark)
					mask |= PF_NOINIT;
				pflg[pc] = pflg[pc] & ~mask | PF_REF;		/* flag reference */
			}

			if (j & OPT_REL)				/* if relative reference */
			{
				if (j & OPT_3)					/* if 3 byte relative address */
				{
					j = pmem[i+2];					/* get offset */
					rel = (j > 0x7f) ? j | 0xff00 : j & 0xff;
					pc = i + 3;
				}
				else								/* else 2 byte relative address */
				{
					j = pmem[i+1];					/* get offset */
					rel = (j > 0x7f) ? j | 0xff00 : j & 0xff;
					pc = i + 2;
				}
				pc += rel;
				if (pc >= offset && pc <= himark)
					mask |= PF_NOINIT;
				pflg[pc] = pflg[pc] & ~mask | PF_REF;		/* flag reference */
			}
			i = i + (opttbl[k] & OPT_SIZE) + 1;		/* update location pointer */
		}
		else										/* not code */
			i++;
		if ((i & 0xff) < l)
			printf("\rPass 1 %04x", i & 0xff00);
	}
	printf("\rPass 1 - Reference search complete");

#ifdef	DEBUG

	dump_pflg();

#endif
}												/*  End of Pass 1 */

/************************************************
 *																*
 *		Output ascii data accumulated in buffer	*
 *																*
 ************************************************/

void dump_ascii(word adrs)
{
	word	padrs, off, cnt;
	char	*cptr;

	padrs = adrs - asccount;		/* print address for comment field */

	/* for test only...
	printf("\rdump_ascii: adrs = 0x%04x, print adrs = 0x%04x, count = %d\n",
		adrs, padrs, asccount);	*/

	adrs = padrs;						/* address in program array */
	cnt = off = 0;						/* cnt = char count, off = buffer offset */
	while (asccount)					/* while data in ascii buffer... */
	{
		if (pflg[adrs] & PF_REF)		/* if addresss is referenced... */
		{
			if (cnt)
			{
				putc('\'', fp);			/* terminate line */
				kcnt++;
				if (hexflag)				/* if comment field requested... */
				{
					do								/* show hex address */
					{
						putc('\t', fp);		/* but tab out to field first */
						kcnt = (kcnt + 8) & 0x78;
					} while (kcnt < xstop);

					fprintf(fp, "; %04x", padrs);
				}
				padrs += cnt;					/* update print address */
				cnt = 0;							/* clear char count for this line */
			}
			cptr = find_entry(adrs, lab_tab);
												/* see if label exists for this adrs */
			if (cptr == NULL)				/* if not, show address in hex */
				fprintf(fp, "\nX%04x:\t%s\t'", adrs, ascistr);
			else								/* else show label name */
				fprintf(fp, "\n%s:\t%s\t'", cptr, ascistr);
			kcnt = 17;
		}
		else if (!cnt)
		{
			fprintf(fp, "\n\t%s\t'", ascistr);
			kcnt = 17;
		}
		putc(string[off], fp);			/* output data in ascii */
		kcnt++;								/* character position in this line */
		cnt++;								/* increment char in this line */
		off++;								/* increment offset into asci buffer */
		adrs++;								/* offset into program memory */
		if (cnt >= ASCLINE)				/* if max characters per line... */
		{
			putc('\'', fp);				/* terminate line */
			kcnt++;
			if (hexflag)					/* if comment field requested */
			{
				do									/* show hex address */
				{
					putc('\t', fp);
					kcnt = (kcnt + 8) & 0x78;
				} while (kcnt < xstop);

				fprintf(fp, "; %04x", padrs);
			}
			padrs += cnt;					/* update print address */
			cnt = 0;
		}
		--asccount;
	}
	putc('\'', fp);					/* terminate line */
	kcnt++;
	if (hexflag && cnt)				/* if comment field requested... */
	{
		do										/* show address */
		{
			putc('\t', fp);
			kcnt = (kcnt + 8) & 0x78;
		} while (kcnt < xstop);

		fprintf(fp, "; %04x", padrs);
	}
	dump = 1;
}

/************************************************
 *																*
 *		Output binary data accumulated in buffer	*
 *																*
 ************************************************/

void dump_bytes(word adrs)
{
	word	padrs, bcnt, off, k;
	char	*cptr;

	padrs = adrs - byte_cnt;			/* compute adrs to print in ascii part */

	/* for test only...
	printf("\rdump_bytes: adrs = 0x%04x, print adrs = 0x%04x, count = %d\n",
		adrs, padrs, byte_cnt);	*/

	adrs = padrs;
	bcnt = off = 0;						/* no bytes output yet */
	while (byte_cnt)						/* while data in binary buffer... */
	{
		if (pflg[adrs] & PF_REF)		/* if data adrs is referenced... */
		{
			if (off && hexflag)				/* dump any remaining ascii first */
			{
				do
				{
					putc('\t', fp);
					kcnt = (kcnt + 8) & 0x78;
				} while (kcnt < xstop);

				fprintf(fp, "; %04x ", padrs);
				for (k=0; k<off; k++)
					putc(ascii(pmem[padrs + k]), fp);
				padrs += k;						/* update print address */
				off = 0;
			}
			cptr = find_entry(adrs, lab_tab);	/* then do a label */
			if (cptr == NULL)
				fprintf(fp, "\nX%04x:\t%s\t", adrs, defbstr);
			else
				fprintf(fp, "\n%s:\t%s\t", cptr, defbstr);
			kcnt = 16;
			bcnt = 0;
		}
		else if (!bcnt)						/* else if first byte... */
		{
			kcnt = 16;
			fprintf(fp, "\n\t%s\t", defbstr);
		}
		else
		{
			putc(',', fp);						/* else separate bytes */
			kcnt++;
		}
		cptr = find_entry(pmem[adrs], sym_tab);
		if (cptr)
			kcnt += fprintf(fp, "%s", cptr);
		else
			puthex(pmem[adrs]);
		bcnt++;
		if (bcnt >= BYTELINE)				/* if max bytes per line... */
		{
			bcnt = 0;
			if (hexflag)						/* do ascii dump of previous bytes */
			{
				do
				{
					putc('\t', fp);
					kcnt = (kcnt + 8) & 0x78;
				} while (kcnt < xstop);

				fprintf(fp, "; %04x ", padrs);
				for (k=0; k<BYTELINE; k++)
					putc(ascii(pmem[padrs + k]), fp);
				padrs += k;
				off = 0;
			}
		}
		else
			off++;
		--byte_cnt;
		adrs++;
	}
	if (off && hexflag)					/* generate comment line */
	{
		do
		{
			putc('\t', fp);
			kcnt = (kcnt + 8) & 0x78;
		} while (kcnt < xstop);

		fprintf(fp, "; %04x ", padrs);	/* show address and ascii for data */
		for (k=0; k<off; k++)
			putc(ascii(pmem[padrs + k]), fp);
	}
	dump = 1;
}

/* end of d51p1.c */

